# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import calendar
import contextlib
from datetime import (
    date,
    datetime,
    timedelta,
)
import logging
import os
import shutil
import subprocess
import sys
import traceback
from types import StringTypes
import urllib2

from bzrlib.branch import Branch
from bzrlib.revision import NULL_REVISION
from bzrlib.transport import get_transport
import requests
import yaml

from charmworld.charmstore import (
    CharmStore,
    get_address,
    make_store_url,
)
from charmworld.lp import (
    get_branch_info,
    parse_date,
)
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.models import getfs
from charmworld.models import CharmFileSet
from charmworld.models import CharmSource
from charmworld.search import (
    ElasticSearchClient,
    SearchServiceNotAvailable,
)
from charmworld.utils import (
    quote_key,
    quote_yaml,
    timestamp,
)

from config import CHARM_DIR
from config import CHARM_PROOF_PATH
from config import settings

ICON_FILENAME = 'icon.svg'
JENKINS_PROVIDERS = ['ec2', 'openstack', 'local']
JENKINS_ARTIFACT_URL = (
    "https://jenkins.qa.ubuntu.com/job/"
    "%(series)s-%(provider)s-charm-%(charm)s/%(build)d/artifact/%(artifact)s")
JENKINS_QA_URL = (
    "https://jenkins.qa.ubuntu.com/job/"
    "%(series)s-%(provider)s-charm-%(charm)s/lastBuild/api/json")


class IngestError(Exception):
    pass


class IngestJob(object):

    name = 'default'

    def __init__(self):
        self.log = logging.getLogger("charm.%s" % self.name)

    def setup(self):
        pass

    def run(self, charm_data):
        raise NotImplemented


class DBIngestJob(IngestJob):

    def setup(self, db=None):
        if not db:
            connection = getconnection(settings)
            db = getdb(connection, settings.get('mongo.database'))
        self.db = db


class FSIngestJob(DBIngestJob):

    def setup(self, db=None, fs=None):
        super(FSIngestJob, self).setup(db)
        if not fs:
            fs = getfs(self.db)
        self.fs = fs


class BzrIngestJob(FSIngestJob):

    name = 'bzr'

    def setup(self, root_dir=None, db=None, fs=None):
        super(BzrIngestJob, self).setup(db, fs)
        if not root_dir:
            root_dir = CHARM_DIR
        self.root_dir = root_dir
        if not os.path.exists(self.root_dir):
            os.makedirs(self.root_dir)

    def store_branch_files(self, charm_data):
        """Process the bzr branch for files that need to be stored in gridfs.
        """
        self.log.info('Storing files of branch into gridfs')
        filestore = CharmFileSet.save_files(
            self.fs, charm_data, charm_data['branch_dir'], self.log)
        self.log.info('Completed gridfs storage.')
        return filestore

    def add_files(self, charm_data):
        charm_data['files'] = dict([
            (quote_key(cfile.filename), dict(cfile)) for cfile in
            self.store_branch_files(charm_data)
        ])
        return charm_data

    def checkout_charm(self, charm_data, branch_dir):
        # The branch has never been seen before. Original branch.
        self.log.info("Branching charm lp:%s", charm_data["branch_spec"])
        subprocess.check_output(
            ["/usr/bin/bzr", "co", "-q",
             "lp:%s" % charm_data["branch_spec"], branch_dir])
        charm_data = self.add_files(charm_data)

    def charm_is_current(self, charm_data, branch_dir):
        # It exists and check if it's the latest revision already.
        self.log.debug(
            "Existing charm from lp:%s", charm_data["branch_spec"])
        transport = get_transport(branch_dir)
        branch = Branch.open_from_transport(transport)
        cur_rev_id = branch.last_revision()
        return cur_rev_id == charm_data['commit']

    def update_charm(self, charm_data, branch_dir, retry=False):
        self.log.debug("Updating branch lp:%s", charm_data["branch_spec"])
        try:
            subprocess.check_output(
                ["/usr/bin/bzr", "update", "-q"],
                cwd=branch_dir,
                stderr=subprocess.STDOUT)
            charm_data = self.add_files(charm_data)
        except subprocess.CalledProcessError:
            # Update failed for some reason; destroy it and start over.
            if retry:
                shutil.rmtree(branch_dir)
                return self.run(charm_data, retry=False)
            raise

    def run(self, charm_data, retry=True):
        """Fetch a branch from bzr, and augment charm data."""
        if charm_data['branch_deleted']:
            return
        branch_dir = os.path.abspath(
            str(os.path.join(self.root_dir,
                             charm_data["series"],
                             charm_data["owner"],
                             charm_data["name"],
                             charm_data["bname"])))

        if not os.path.exists(os.path.dirname(branch_dir)):
            os.makedirs(os.path.dirname(branch_dir))
        # Store the branch directory
        charm_data["branch_dir"] = branch_dir

        if not os.path.exists(branch_dir):
            # Charm doesn't exist; check it out.
            self.checkout_charm(charm_data, branch_dir)
            return
        elif self.charm_is_current(charm_data, branch_dir):
            # Charm exists, and is current; log and finish.
            charm_data = self.add_files(charm_data)
            self.log.debug(
                "Already up to date lp:%s", charm_data["branch_spec"])
            return
        else:
            # Charm exists, but needs updating; update it.
            self.update_charm(charm_data, branch_dir, retry)


def log(stage, level, exc, charm_data, tb=None):
    logger = logging.getLogger("charm.%s" % stage)
    err_msg = "%s error on %s: %s" % (
        stage, charm_data, str(exc))
    if tb is not None:
        err_msg = '%s\n%s' % (err_msg, tb)
    logger.log(level, err_msg)
    charm_error = {'error_stage': stage, 'error': str(exc)}
    if hasattr(exc, 'output'):
        charm_error['output'] = str(exc.output)
    charm_data['error'] = charm_error


def run_job(job, charm_data, needs_setup=True, db=None):
    if needs_setup:
        if db is not None:
            job.setup(db=db)
        else:
            job.setup()
    stage = job.name
    try:
        job.run(charm_data)
    except IngestError as e:
        log(stage, logging.INFO, e, charm_data)
        return False
    except SearchServiceNotAvailable:
        raise
    except Exception as e:
        tb = traceback.format_exc()
        log(stage, logging.ERROR, e, charm_data, tb)
        return False
    return True


def update_charm(charm_data, db):
    charm_data["_id"] = charm_data["branch_spec"]
    # Drop existing error data so that charms can lose their error status.
    charm_data.pop('error', None)
    log = logging.getLogger("charm.update_charm")
    try:
        if not run_job(BzrIngestJob(), charm_data, db=db):
            return False
        if not run_job(StoreIngestJob(), charm_data):
            return False
        if not run_job(ProofIngestJob(), charm_data):
            return False
        if not run_job(JenkinsIngestJob(), charm_data):
            return False
        if not run_job(ChangelogIngestJob(), charm_data):
            return False
        update_date_created(charm_data, log)
        if not run_job(ScanIngestJob(), charm_data, db=db):
            return False
    except Exception as e:
        err_msg = "%s error: %s" % (charm_data, str(e))
        log.exception(err_msg)
        charm_error = {'error': str(e)}
        if hasattr(e, 'output'):
            charm_error['output'] = str(e.output)
        charm_data['error'] = charm_error


class UpdateCharmJob(DBIngestJob):

    name = 'update'

    def run(self, payload):
        charm_data = self.db.charms.find_one(payload['branch_spec'])
        if charm_data is None:
            charm_data = {}
        charm_data.update(payload)
        update_charm(charm_data, self.db)
        index_client = ElasticSearchClient.from_settings(settings)
        self.log.info('Saving %s' % charm_data['branch_spec'])
        CharmSource(self.db, index_client).save(charm_data)


class ChangelogIngestJob(IngestJob):

    name = 'changelog'

    def __init__(self, limit=10, since=None):
        self.limit = limit
        self.since = since

    def setup(self):
        super(ChangelogIngestJob, self).setup()
        if self.since is None:
            days_of_revisions = settings.get('days_of_revisions')
            if days_of_revisions is not None:
                cutoff = datetime.utcnow() - timedelta(int(days_of_revisions))
                self.since = calendar.timegm(cutoff.timetuple())

    @staticmethod
    def _rev_info(r, branch):
        d = {
            'authors': r.get_apparent_authors(),
            "revno": branch.revision_id_to_revno(r.revision_id),
            "committer": r.committer,
            "created": r.timestamp,
            "message": r.message
        }
        return d

    def run(self, charm_data):
        if charm_data['branch_deleted']:
            return
        branch_dir = charm_data["branch_dir"]
        charm_data.update(self.get_changes(branch_dir))

    def get_changes(self, branch_dir):
        charm_data = {}
        branch = Branch.open(branch_dir)
        branch.lock_read()
        try:
            revisions = self.get_revisions(branch)
            charm_data["changes"] = changes = []
            for r in revisions:
                changes.append(self._rev_info(r, branch))
            if len(revisions) == 0:
                last_change = None
                first_change = None
            else:
                last_change = changes[0]
                first = branch.repository.get_revision(branch.get_rev_id(1))
                first_change = self._rev_info(first, branch)
            charm_data.update({
                'last_change': last_change,
                'first_change': first_change,
            })
            return charm_data
        finally:
            branch.unlock()

    def get_revisions(self, branch):
        # We only want the last 10 changes, in descending order.
        graph = branch.repository.get_graph()
        cur_rev_id = branch.last_revision()
        ancestry_iter = graph.iter_lefthand_ancestry(cur_rev_id)
        revs = []
        for num, revision_id in enumerate(ancestry_iter):
            if revision_id == NULL_REVISION:
                break
            revision = branch.repository.get_revision(revision_id)
            if num >= self.limit:
                if self.since is None or revision.timestamp < self.since:
                    break
            revs.append(revision)
        return revs


class JenkinsIngestJob(FSIngestJob):

    name = 'jenkins'

    def run(self, charm):
        if not charm['promulgated']:
            return

        charm['tests'] = {}
        charm['test_results'] = {}

        for p in JENKINS_PROVIDERS:
            try:
                result_id, status = self.store_provider_results(p, charm)
                if result_id is None:
                    continue
                charm['tests'][p] = status
                charm['test_results'][p] = result_id
            except:
                self.log.exception("Unknown error while processing %s %s",
                                   charm['branch_spec'], p)

    def _fetch_artifacts(self, provider, charm, result):
        artifacts = []
        for artifact in result['artifacts']:
            a_path = "%s/%s/%s/%s" % (
                charm['branch_spec'],
                provider,
                result['number'],
                artifact['displayPath'])
            a_url = JENKINS_ARTIFACT_URL % (dict(
                series=charm['series'],
                provider=provider,
                charm=charm['name'],
                build=result['number'],
                artifact=artifact['relativePath']))

            ## File sniffing

            # Load up the charm revision as a result property.
            if artifact['displayPath'] == 'charm-revision':
                charm_revision = urllib2.urlopen(a_url).read().strip()
                if not charm_revision:
                    continue
                revision_info = charm_revision.split(' ', 1)
                result['revno'] = int(revision_info[0])
                if len(revision_info) > 1:
                    result['revision-id'] = revision_info[1].rstrip('\n')
                continue

            # Mark the test result as graph runner enabled.
            if "graph-tests" in artifact['displayPath']:
                result['charmrunner'] = True

            # Skip the actual charm content
            if "charm-%s.zip" % charm['name'] == artifact['displayPath']:
                continue

            # XXX Short circuit before actual fetching.
            continue

            a_file = urllib2.urlopen(a_url)
            file_id = self.fs.put(a_file, path=a_path)
            artifact['file_id'] = file_id
            artifacts.append(artifact)

        return artifacts

    def store_provider_results(self, provider, charm):
        charm_result_url = JENKINS_QA_URL % (
            dict(series=charm['series'],
                 provider=provider,
                 charm=charm['name']))

        self.log.debug("Loading %s from %s", charm['name'], charm_result_url)

        response = requests.get(charm_result_url)
        if response.status_code == 404:
            self.log.debug(
                "No test result for %s @ %s", charm['branch_spec'], provider)
            return None, None

        result = response.json()

        # If we already have results no pointing in refetching.
        result_id = "%s::%s-%s" % (
            charm['branch_spec'], provider, result['number'])
        db_result = self.db.jenkins.find_one({'_id': result_id})
        if db_result is not None:
            return result_id, db_result['result']

        # Fetch test artifacts.
        artifacts = self._fetch_artifacts(provider, charm, result)

        # Inject test metadata.
        result['branch_spec'] = charm['branch_spec']
        result['provider'] = provider
        result['artifacts'] = artifacts
        result['_id'] = result_id
        self.db.jenkins.insert(result)
        return (result_id, result['result'])


class ProofIngestJob(IngestJob):

    name = 'proof'

    @contextlib.contextmanager
    def _get_proof_lib(self, new_path):
        if new_path not in sys.path:
            sys.path.append(new_path)
        try:
            import lib.proof as prooflib
            yield prooflib
        except ImportError:
            yield None
        finally:
            if new_path in sys.path:
                sys.path.remove(new_path)

    def setup(self, proofer=None, proof_path=None):
        if not proofer:
            proofer = self.get_proofer(proof_path)
        self.proofer = proofer

    def get_proofer(self, proof_path=None):
        proofer = None
        if not proof_path:
            proof_path = CHARM_PROOF_PATH

        # Use config.CHARM_PROOF_PATH for testing. Monkeypatching is scary.
        if not os.path.isdir(proof_path):
            err_msg = ("proof error before processing began: could not find "
                       "charm proof path.")
            self.log.exception(err_msg)
            self.log.exception("CHARM_PROOF_PATH: %s", proof_path)
        with self._get_proof_lib(proof_path) as prooflib:
            if not prooflib:
                err_msg = ("proof error before processing began: could not "
                           "import charm proof lib.")
                self.log.exception(err_msg)
                self.log.exception(
                    "CHARM_PROOF_PATH: %s", proof_path)
            else:
                proofer = prooflib.run
        return proofer

    def run(self, charm):
        if not self.proofer:
            self.log.exception("proof aborted.")
            raise Exception("No proofer")
        if charm['branch_deleted']:
            return
        proof = {}
        lint, exit_code = self.proofer(charm['branch_dir'])
        for line in lint:
            if not ':' in line:
                continue
            level, msg = line.split(':', 1)
            if level == "W" and 'name' in msg:
                continue
            proof.setdefault(level.lower(), []).append(msg)
        charm['proof'] = proof


class ScanIngestJob(FSIngestJob):

    name = 'scan'

    @staticmethod
    def normalize_interfaces(interfaces):
        # Transform the short defintion of relations into the
        # long form. Quoted from
        # https://juju.ubuntu.com/docs/charm.html#the-metadata-file :
        #
        #     As a shortcut, if these properties are not defined, and
        #     instead a single string value is provided next to the
        #     relation name, the string is taken as the interface
        #     value, as seen in this example:
        #
        #         requires:
        #             db: mysql
        for relation_name, relation_data in interfaces.items():
            if isinstance(relation_data, StringTypes):
                interfaces[relation_name] = {'interface': relation_data}

    @classmethod
    def process_charm(cls, charm):
        # Enrich charm metadata for webapp.

        # Charm url
        if charm["promulgated"]:
            charm["short_url"] = "/charms/%s/%s" % (
                charm["series"], charm["name"])
        else:
            charm["short_url"] = "/~%s/%s/%s" % (charm["owner"],
                                                 charm["series"],
                                                 charm["name"])

        # Charm label
        if charm["promulgated"]:
            charm["label"] = "%s/%s" % (charm["series"], charm["name"])
        else:
            charm["label"] = "~%s:%s/%s" % (charm["owner"],
                                            charm["series"],
                                            charm["name"])

        # Flatten the interfaces provided
        i_provides = []
        provides = charm.get("provides")
        if provides:
            cls.normalize_interfaces(provides)
            for v in provides.values():
                if not isinstance(v, dict):
                    continue
                i = v.get("interface")
                if not i:
                    continue
                i_provides.append(i)
        charm["i_provides"] = i_provides

        # Flatten the interfaces required
        i_requires = []
        requires = charm.get("requires")
        if requires:
            cls.normalize_interfaces(requires)
            for v in requires.values():
                i = v.get("interface")
                if not i:
                    continue
                i_requires.append(i)
        charm["i_requires"] = i_requires
        return charm

    def run(self, charm_data):
        # Note: charm_data is modified in-place.  IndexIngestJob requires
        # these modifications.
        files = charm_data['files']
        # Some files have bad characters in them since they are used as mongo
        # keys. Use their escaped forms instead.
        metadata_file = quote_key('metadata.yaml')
        config_file = quote_key('config.yaml')

        if metadata_file not in files:
            self.log.info(
                "Charm has no metadata: %s", charm_data["branch_spec"])
            return
        else:
            cfile = CharmFileSet.get_by_id(
                self.fs, files[metadata_file]['fileid'])
            try:
                metadata = quote_yaml(yaml.load(cfile.read()))
            except Exception, exc:
                raise IngestError(
                    'Invalid charm metadata %s: %s' % (
                    charm_data['branch_spec'],
                    exc)
                )

        if config_file in files:
            cfile = CharmFileSet.get_by_id(
                self.fs, files[config_file]['fileid'])
            config_raw = cfile.read()

            try:
                config = quote_yaml(yaml.load(config_raw))
            except Exception, exc:
                raise IngestError(
                    'Invalid charm config yaml. %s: %s' % (
                        charm_data['branch_spec'],
                        exc)
                )

            metadata["config"] = config
            metadata["config_raw"] = config_raw

        if 'revision' in files:
            cfile = CharmFileSet.get_by_id(
                self.fs, files['revision']['fileid'])
            rev_raw = cfile.read()
            rev_id = int(rev_raw.strip())
            metadata["revision"] = rev_id
        elif not "revision" in metadata:
            self.log.info("Invalid revision %s", charm_data["branch_spec"])
            metadata["revision"] = 0

        hooks = []
        for filedata in files.values():
            if filedata['subdir'] == 'hooks':
                hooks.append(filedata['filename'])
        hooks.sort()
        metadata["hooks"] = hooks

        # Stuff into the db
        metadata.update(charm_data)
        metadata = self.process_charm(metadata)
        # Modify charm_data itself.
        charm_data.clear()
        charm_data.update(metadata)


def addresses(charm):
    """Return an iterator of potential store addresses."""
    if charm['promulgated']:
        yield get_address(charm, short=True)
    yield get_address(charm, short=False)


def update_download_count(store, charm):
    count = store.count_downloads_in_days(charm, 30, date.today())
    charm['downloads_in_past_30_days'] = count


def update_date_created(charm, log):
    """Use Launchpad to update the charm's date_created field.

    Field will only be set if it is missing AND a branch with the correct
    branch_spec exists in Launchpad.
    """
    if 'date_created' in charm:
        log.debug('Skipping %s which already has date_created',
                  charm['branch_spec'])
        return
    log.info('Retrieving date_created for %s', charm['branch_spec'])
    info = get_branch_info(charm['branch_spec'])
    if info is None:
        log.warning('No branch for %s', charm['branch_spec'])
        return
    date_created = parse_date(info['date_created'])
    charm['date_created'] = timestamp(date_created.replace(microsecond=0))


class StoreIngestJob(IngestJob):

    name = 'store'

    def __init__(self, store=None):
        super(StoreIngestJob, self).__init__()
        if store is None:
            store = CharmStore()
        self.store = store

    def run(self, charm):
        old_address = None
        for address in addresses(charm):
            if old_address is not None:
                self.log.info("rechecking %s with %s", old_address, address)
            old_address = address
            data = self._store_get(address)
            if 'errors' not in data and 'warnings' not in data:
                break

        if 'errors' in data or 'warnings' in data:
            self.log.warning("store error on %s %s" % (address, data))

        data["store_checked"] = datetime.now().ctime()

        charm['store_data'] = data
        charm['store_url'] = make_store_url(data['revision'], address)
        update_download_count(self.store, charm)

    def _store_get(self, address):
        return CharmStore().get_charm_info(address)


# XXX j.c.sackett Jan 31 2013 Bug:1111708 scan_repo is swapped for
# scan_charm to reindex what's on disk; we should probably just
# have a different job for this that's not part of ingest.
#def scan_repo(db, root_dir):
    #log = logging.getLogger("charm.scan")
    #charms = os.listdir(root_dir)
    #for c in charms:
        #charm_dir = os.path.join(root_dir, c)
        #if not os.path.isdir(charm_dir):
            #continue
        ##log.info("Processing %s", c)
        #try:
            #scan_charm(
                #db, c, charm_dir, repo="~charmers/charm/oneiric/%s" % c)
        #except:
            #log.exception("Unknown scan error")
            #raise
            #import pdb
            #import sys
            #import traceback
            #traceback.print_exc()
            #pdb.post_mortem(sys.exc_info()[-1])
            #raise
