# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

import argparse
import logging
import pymongo

from charmworld.lp import get_branch_tips
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.utils import (
    configure_logging,
    get_ini,
)
from config import CHARM_IMPORT_FILTER
from config import CHARM_QUEUE
from utils import get_queue
from utils import lock
from utils import LockHeld
from utils import parse_branch


default = object()


def available_charms(charm_data=None):
    # Return data needed for the ingest job for charm branches on Launchpad.
    log = logging.getLogger("charm.launchpad")
    if not charm_data:
        charm_data = get_branch_tips()

    for repo, commit, series in charm_data:
        _, branch_name = repo.rsplit("/", 1)

        try:
            data = parse_branch(repo, series, commit)
        except ValueError:
            log.warning("Unable to parse repo %s", repo)
            continue

        if data["bname"] not in ("trunk", "trunk-1"):
            log.debug("Skipped branch %s", repo)
            continue

        data['promulgated'] = data['series'] in data['distro_series']
        data['branch_deleted'] = False

        yield data


def db_charms(db):
    # Return data needed for the ingest job for charms stored in
    # charmworld's database.
    for charm in db.charms.find(
            fields=('owner', 'series', 'name', 'bname', 'branch_spec'),
            sort=[('branch_spec', pymongo.ASCENDING)]):
        # Override the promulgation related settings: Data from this
        # iterator is used by all_charms() only when the charm has no
        # longer a related Launchpad branch. A charm is promulgated if
        # its LP branch is linked to a sourcepackage, hence a charm
        # without a branch cannot be promulgated.
        charm['promulgated'] = False
        charm['distro_series'] = []
        charm['branch_deleted'] = True
        yield charm


def all_charms(db, charm_data=None, limit=None, import_filter=default):
    # Return a sequence of charms that have a Launchpad branch or that are
    # already known in charmworld's database.
    log = logging.getLogger("charm.launchpad")
    if import_filter is default:
        import_filter = CHARM_IMPORT_FILTER
    all = dict((charm['branch_spec'], charm) for charm in db_charms(db))
    available = available_charms(charm_data)
    all.update((charm['branch_spec'], charm) for charm in available)
    all = sorted(all.values(), key=lambda charm: charm['branch_spec'])

    count = 0
    for charm in all:
        if import_filter and not charm['branch_spec'].startswith(
                import_filter):
            continue
        log.info("Queueing %s", charm['branch_spec'])
        yield charm
        count += 1
        if limit and count >= limit:
            log.info("Import limit reached (%d)... stopping", limit)
            break


def queue_charms(db, out_queue, limit=None, import_filter=default):
    log = logging.getLogger("charm.launchpad")
    for charm in all_charms(db, limit=limit, import_filter=import_filter):
        added = out_queue.put(charm)
        if added and 0:
            log.info("Queued %s", charm)


def main():
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--limit",
        help="Maximum number of charms to queue.",
        type=int)
    parser.add_argument(
        "--prefix",
        help="Only charms matching this prefix are queued.",
        type=str, default=default)
    args = parser.parse_args()
    log = logging.getLogger("charm.launchpad")
    settings = get_ini()
    if args.limit is None:
        charm_import_limit = settings.get('charm_import_limit')
        charm_import_limit = (
            int(charm_import_limit) if charm_import_limit else None)
    else:
        charm_import_limit = args.limit
    connection = getconnection(settings)
    db = getdb(connection, settings.get('mongo.database'))
    try:
        with lock('ingest-queue', int(settings['script_lease_time']) * 60,
                  db, log):
            queue = get_queue(CHARM_QUEUE)
            queue_charms(db, queue, charm_import_limit, args.prefix)
    except LockHeld, error:
        log.warn(str(error))


if __name__ == '__main__':
    main()
