# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).


from contextlib import contextmanager
import os
import shutil

from bzrlib.bzrdir import BzrDir
from mock import patch


from charmworld.jobs.ingest import (
    BzrIngestJob,
    ChangelogIngestJob,
)
from charmworld.testing import (
    factory,
    JobTestBase,
    TestCase,
    temp_dir,
)


JRH = 'J. Random Hacker <jrandom@example.com>'
YRH = 'Y. Random Hacker <yrandom@example.com>'
ZRH = 'Z. Random Hacker <zrandom@example.com>'


class TestBzrJob(JobTestBase):

    def setUp(self):
        super(TestBzrJob, self).setUp()
        self.test_dir = 'test_bzr_job'
        os.mkdir(self.test_dir)

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        super(TestBzrJob, self).tearDown()

    def test_run_creates_charm_directory(self):
        job = BzrIngestJob()
        job.setup(root_dir=self.test_dir, db=self.db)
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        with patch.object(job, 'checkout_charm'):
            job.run(charm)
        expected_path = os.path.abspath(os.path.join(
            self.test_dir, charm['series'], charm['owner'], charm['name']))
        self.assertTrue(os.path.exists(expected_path))

    def test_run_checkouts_new_charms(self):
        job = BzrIngestJob()
        job.setup(root_dir=self.test_dir, db=self.db)
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        with patch.object(job, 'checkout_charm') as mock:
            job.run(charm)
        self.assertTrue(mock.called)

    def test_run_updates_exising_charms_if_not_current(self):
        job = BzrIngestJob()
        job.setup(root_dir=self.test_dir, db=self.db)
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        os.makedirs(charm_path)
        with patch.object(job, 'charm_is_current', lambda x, y: False):
            with patch.object(job, 'update_charm') as mock:
                job.run(charm)
        self.assertTrue(mock.called)

    def test_run_does_not_update_exising_charms_if_current(self):
        job = BzrIngestJob()
        job.setup(root_dir=self.test_dir, db=self.db)
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        os.makedirs(charm_path)
        with patch.object(job, 'charm_is_current', lambda x, y: True):
            with patch.object(job, 'add_files'):
                with patch.object(job, 'update_charm') as mock:
                    job.run(charm)
        self.assertFalse(mock.called)

    def test_add_files(self):
        charm_data = {
            'branch_dir': os.path.join(
                os.path.dirname(__file__), '../../testing/data/sample_charm'),
            'owner': 'charmers',
            'series': 'precise',
            'name': 'sample_charm',
            'bname': 'trunk',
        }
        job = BzrIngestJob()
        job.setup(db=self.db)
        job.add_files(charm_data)
        self.assertIn('files', charm_data)
        self.assertNotIn('icon', charm_data)

    def test_deleted_branch(self):
        # If the Launchpad branch of a charm is deleted, BzrJob does
        # not do anything.
        job = BzrIngestJob()
        job.setup(root_dir=self.test_dir, db=self.db)
        ignore, charm = factory.makeCharm(
            self.db, branch_root=self.test_dir, branch_deleted=True)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        with patch.object(job, 'checkout_charm') as checkout_mock:
            with patch.object(job, 'charm_is_current') as is_current_mock:
                with patch.object(job, 'add_files') as add_files_mock:
                    with patch.object(job, 'update_charm') as update_mock:
                        job.run(charm)
        self.assertFalse(os.path.exists(charm_path))
        self.assertFalse(checkout_mock.called)
        self.assertFalse(is_current_mock.called)
        self.assertFalse(add_files_mock.called)
        self.assertFalse(update_mock.called)


@contextmanager
def bzr_isolation():
    with temp_dir() as temp_home:
        old_cwd = os.getcwd()
        os.chdir(temp_home)
        try:
            with patch.dict(os.environ, values={'HOME': temp_home},
                            clear=True):
                yield temp_home
        finally:
            os.chdir(old_cwd)


@contextmanager
def write_locked(lockable):
    lockable.lock_write()
    try:
        yield lockable
    finally:
        lockable.unlock()


class TestChangelogBzrHandling(TestCase):

    def setUp(self):
        super(TestChangelogBzrHandling)
        self.use_context(bzr_isolation())

    @staticmethod
    def num_commit(tree, num):
        return tree.commit("revision %d" % num, committer=JRH)

    def make_locked_tree(self):
        tree = BzrDir.create_standalone_workingtree('tree')
        return self.use_context(write_locked(tree))

    def test_get_revisions(self):
        tree = self.make_locked_tree()
        job = ChangelogIngestJob()
        revisions = job.get_revisions(tree.branch)
        self.assertEqual([], revisions)
        for x in range(9):
            self.num_commit(tree, x)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(["revision %d" % num for num in range(8, -1, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 9)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(["revision %d" % num for num in range(9, -1, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 10)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(["revision %d" % num for num in range(10, 0, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 11)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(["revision %d" % num for num in range(11, 1, -1)],
                         [revision.message for revision in revisions])

    def test_get_revisions_honours_limit(self):
        tree = self.make_locked_tree()
        for x in range(15):
            self.num_commit(tree, x)
        revisions = ChangelogIngestJob(limit=5).get_revisions(tree.branch)
        self.assertEqual(5, len(revisions))
        revisions = ChangelogIngestJob(limit=20).get_revisions(tree.branch)
        self.assertEqual(15, len(revisions))

    def test_get_revisions_honours_since(self):
        tree = self.make_locked_tree()
        tree.commit('', timestamp=0, committer='')
        tree.commit('', timestamp=1, committer='')
        tree.commit('', timestamp=2, committer='')
        job = ChangelogIngestJob(since=0, limit=0)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(3, len(revisions))
        job = ChangelogIngestJob(since=1, limit=0)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(2, len(revisions))

    def test_limit_overrides_since(self):
        tree = self.make_locked_tree()
        tree.commit('', timestamp=0, committer='')
        tree.commit('', timestamp=1, committer='')
        tree.commit('', timestamp=2, committer='')
        job = ChangelogIngestJob(since=1, limit=3)
        revisions = job.get_revisions(tree.branch)
        self.assertEqual(3, len(revisions))

    def test_rev_info(self):
        tree = self.make_locked_tree()
        revision_id = tree.commit('message1', committer=JRH)
        rev = tree.branch.repository.get_revision(revision_id)
        rev.timestamp = 34
        rev_info = ChangelogIngestJob._rev_info(rev, tree.branch)
        self.assertEqual({
            'authors': [JRH],
            'committer': JRH,
            'created': 34,
            'message': 'message1',
            'revno': 1,
        }, rev_info)

    @staticmethod
    def get_rev_info(branch, revision_id):
        rev = branch.repository.get_revision(revision_id)
        return ChangelogIngestJob._rev_info(rev, branch)

    def test_rev_info_supplied_authors(self):
        tree = self.make_locked_tree()
        revision_id = tree.commit(
            'message1', committer=JRH, authors=[YRH, ZRH])
        rev_info = self.get_rev_info(tree.branch, revision_id)
        self.assertEqual([YRH, ZRH], rev_info['authors'])
        self.assertEqual(JRH, rev_info['committer'])

    def test_get_changes(self):
        tree = self.make_locked_tree()
        first_id = self.num_commit(tree, 1)
        job = ChangelogIngestJob()
        charm_data = job.get_changes('tree')
        self.assertItemsEqual(['changes', 'first_change', 'last_change'],
                              charm_data.keys())
        self.assertEqual(charm_data['first_change'], charm_data['last_change'])
        self.assertEqual([charm_data['first_change']], charm_data['changes'])
        for num in range(1, 11):
            self.num_commit(tree, 1)
        charm_data = job.get_changes('tree')
        self.assertNotEqual(charm_data['first_change'],
                            charm_data['last_change'])
        self.assertEqual(charm_data['last_change'], charm_data['changes'][0])
        self.assertNotEqual(charm_data['first_change'],
                            charm_data['changes'][-1])
        self.assertEqual(self.get_rev_info(tree.branch, first_id),
                         charm_data['first_change'])

    def test_get_changes_no_revisions(self):
        self.make_locked_tree()
        charm_data = ChangelogIngestJob().get_changes('tree')
        self.assertEqual([], charm_data['changes'])
        self.assertIs(None, charm_data['first_change'])
        self.assertIs(None, charm_data['last_change'])

    def test_branch_deleted(self):
        # ChangelogIngestJob does not do anything if the Launchpad branch
        # of a charm is deleted.
        job = ChangelogIngestJob()
        charm = factory.get_charm_json(branch_deleted=True)
        with patch.object(job, 'get_changes') as mock:
            job.run(charm)
        self.assertFalse(mock.called)
