# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from contextlib import contextmanager
import json
import logging
import os.path

from bzrlib.bzrdir import BzrDir
from httmock import (
    HTTMock,
    urlmatch,
)
from mock import patch
from pyelasticsearch.exceptions import ElasticHttpNotFoundError

from charmworld.charmstore import get_address
from charmworld.jobs.ingest import (
    IngestError,
    IngestJob,
    run_job,
    update_charm,
    UpdateCharmJob,
    update_date_created,
)
from charmworld.jobs.tests.test_bzr import bzr_isolation
from charmworld.models import CharmSource
from charmworld.testing import (
    factory,
    force_index_client,
    MongoTestBase,
    response404,
    TestCase,
)


class TestUpdateDateCreated(TestCase):

    def test_update_date_created(self):
        result = json.dumps(
            {'date_created': '2012-03-06T17:59:55.818912+00:00'})
        charm = {'branch_spec': 'steve'}
        with HTTMock(lambda x, y: result):
            update_date_created(charm, logging.getLogger('foo'))
        self.assertEqual({
            'date_created': '2012-03-06T17:59:55Z',
            'branch_spec': 'steve',
        }, charm)

    def test_update_date_created_ignores_existing(self):
        charm = {'date_created': None, 'branch_spec': 'steve'}
        update_date_created(charm, logging.getLogger('foo'))
        self.assertEqual({'date_created': None, 'branch_spec': 'steve'}, charm)

    def test_update_date_created_no_branch(self):
        with HTTMock(response404):
            charm = {'branch_spec': 'steve'}
            update_date_created(charm, logging.getLogger('foo'))
        self.assertNotIn('date_created', charm)


@contextmanager
def update_environment(real_charm_data, index_client):
    def checkout_charm(self, charm_data, branch_dir):
        wt = BzrDir.create_standalone_workingtree(branch_dir)
        wt.bzrdir.root_transport.mkdir('hooks')
        wt.bzrdir.root_transport.put_bytes('metadata.yaml', json.dumps({
            'name': charm_data['name'],
            'summary': real_charm_data['summary'],
        }))
        self.add_files(charm_data)
    checkout_patch = patch(
        'charmworld.jobs.ingest.BzrIngestJob.checkout_charm', checkout_charm)

    @urlmatch(path='/charm-info')
    def mock_store_data(url, request):
        address = get_address(real_charm_data, real_charm_data['promulgated'])
        return json.dumps({address: {'revision': 1}})

    @urlmatch(path='/stats/counter/charm-bundle:.*')
    def mock_counts(url, request):
        return json.dumps([[0]])

    @urlmatch(netloc='api.launchpad.net')
    def mock_date_created(url, request):
        return json.dumps({'date_created': '2005-06-07T12:13:14.15+00:00'})

    test_results = urlmatch(netloc='jenkins.qa.ubuntu.com')(response404)
    httmock = HTTMock(mock_store_data, mock_counts, test_results,
                      mock_date_created)
    with bzr_isolation() as cwd, checkout_patch:
        root_dir = os.path.join(cwd, 'root_dir')
        dir_patch = patch('charmworld.jobs.ingest.CHARM_DIR', root_dir)
        with dir_patch, httmock, force_index_client(index_client):
            yield


class TestUpdateCharmJob(MongoTestBase):

    @staticmethod
    def payload(charm_data):
        return {
            'series': charm_data['series'],
            'owner': charm_data['owner'],
            'name': charm_data['name'],
            'bname': charm_data['bname'],
            'branch_spec': charm_data['branch_spec'],
            'branch_deleted': charm_data['branch_deleted'],
            'promulgated': charm_data['promulgated'],
        }

    def test_run_indexes(self):
        charm_data = factory.get_charm_json()
        with update_environment(charm_data, self.use_index_client()):
            run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        self.assertIsNot(None, self.index_client.get(charm_data['_id']))

    def test_run_updates_mongo(self):
        charm_data = factory.get_charm_json()
        with update_environment(charm_data, self.use_index_client()):
            run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        self.assertIsNot(None, self.db.charms.find_one(charm_data['_id']))

    def test_run_includes_old_mongo_data_in_index(self):
        charm_data = factory.get_charm_json()
        charm_data['foo'] = 'bar'
        self.db.charms.save(charm_data)
        del charm_data['foo']
        with update_environment(charm_data, self.use_index_client()):
            run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        indexed_charm = self.index_client.get(charm_data['_id'])
        self.assertEqual('bar', indexed_charm.get('foo'))

    def test_mongo_and_index_store_identical_charm(self):
        charm_data = factory.get_charm_json()
        charm_data['foo'] = 'bar'
        self.db.charms.save(charm_data)
        del charm_data['foo']
        with update_environment(charm_data, self.use_index_client()):
            run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        indexed_charm = self.index_client.get(charm_data['_id'])
        mongo_charm = self.db.charms.find_one(charm_data['_id'])
        self.assertEqual(indexed_charm, mongo_charm)

    def test_saves_early_error(self):
        charm_data = factory.get_charm_json()
        # Induce improbable error -- bname must be a string.
        charm_data['bname'] = None
        update_handler = self.get_handler('charm.update')
        bzr_handler = self.get_handler('charm.bzr')
        run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        self.assertIsNot(
            None, self.db.charms.find_one(charm_data['_id']))
        logs = ''.join(r.getMessage() for r in update_handler.buffer)
        bzr_logs = ''.join(r.getMessage() for r in bzr_handler.buffer)
        self.assertEqual('Saving %s' % charm_data['_id'], logs)
        self.assertIn("'NoneType' object has no attribute 'startswith'",
                      bzr_logs)

    def test_updates_promulgated(self):
        charm_data = factory.get_charm_json(promulgated=False)
        self.use_index_client()
        CharmSource.from_request(self).save(charm_data)
        charm_data['promulgated'] = True
        with update_environment(charm_data, self.index_client):
            run_job(UpdateCharmJob(), self.payload(charm_data), db=self.db)
        self.assertTrue(
            self.db.charms.find_one(charm_data['_id'])['promulgated'])


class TestUpdateCharm(MongoTestBase):

    def test_update_charm_does_not_index(self):
        charm_data = factory.get_charm_json()
        with update_environment(charm_data, self.use_index_client()):
            update_charm(charm_data, self.db)
        self.index_client.wait_for_startup()
        with self.assertRaises(ElasticHttpNotFoundError):
            self.index_client.get(charm_data['_id'])

    def test_update_charm_does_not_update_mongo(self):
        charm_data = factory.get_charm_json()
        with update_environment(charm_data, self.use_index_client()):
            update_charm(charm_data, self.db)
        self.assertIs(None, self.db.charms.find_one(charm_data['_id']))

    def test_forgets_past_error(self):
        charm_data = factory.get_charm_json(promulgated=True)
        charm_data['error'] = {'error_stage': 'foo', 'error': 'bar'}
        with update_environment(charm_data, self.use_index_client()):
            update_charm(charm_data, self.db)
        self.assertNotIn('error', charm_data)


class TestJob(IngestJob):
    name = 'test'

    def __init__(self, exc=None):
        super(TestJob, self).setup()
        self.setup_called = False
        self.run_called = False
        self.exc = exc
        self.charm_data = None

    def setup(self, db=None):
        self.setup_called = True
        self.db = db

    def run(self, charm_data):
        self.charm_data = charm_data
        if self.exc is not None:
            raise self.exc('test error')
        self.run_called = True


class TestRunJob(TestCase):

    def test_run_job_defaults(self):
        job = TestJob()
        result = run_job(job, charm_data='foo')
        self.assertTrue(result)
        self.assertTrue(job.setup_called)
        self.assertTrue(job.run_called)
        self.assertEqual('foo', job.charm_data)
        self.assertIs(None, job.db)

    def test_run_job_no_setup(self):
        job = TestJob()
        run_job(job, 'foo', needs_setup=False)
        self.assertFalse(job.setup_called)

    def test_run_job_db_specified(self):
        job = TestJob()
        run_job(job, 'foo', db='db')
        self.assertEqual('db', job.db)
        job = TestJob()
        run_job(job, 'foo', needs_setup=False, db='db')
        self.assertIs(None, getattr(job, 'db', None))

    def test_run_job_general_exception(self):
        handler = self.get_handler('charm.test')
        job = TestJob(ValueError)
        charm_data = {'foo': 'bar'}
        result = run_job(job, charm_data)
        self.assertFalse(result)
        expected_error = {
            'error': 'test error',
            'error_stage': 'test',
        }
        self.assertEqual(expected_error, charm_data['error'])
        log_messages = [record.getMessage() for record in handler.buffer]
        expected = "test error on {'foo': 'bar'}: test error\nTraceback"
        self.assertEqual(1, len(log_messages))
        self.assertTrue(log_messages[0].startswith(expected))
        log_level = handler.buffer[0].levelname
        self.assertEqual('ERROR', log_level)

    def test_run_job_IngestError(self):
        handler = self.get_handler('charm.test')
        job = TestJob(IngestError)
        charm_data = {'foo': 'bar'}
        result = run_job(job, charm_data)
        self.assertFalse(result)
        expected_error = {
            'error': 'test error',
            'error_stage': 'test',
        }
        self.assertEqual(expected_error, charm_data['error'])
        log_messages = [record.getMessage() for record in handler.buffer]
        expected = ["test error on {'foo': 'bar'}: test error"]
        self.assertEqual(expected, log_messages)
        log_level = handler.buffer[0].levelname
        self.assertEqual('INFO', log_level)
