# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json

from httmock import HTTMock
from mock import patch

from charmworld.testing import (
    factory,
    JobTestBase,
)

from charmworld.jobs.lp import (
    all_charms,
    db_charms,
    available_charms,
    queue_charms,
)
from charmworld.jobs.utils import get_queue


def available_charms_mock(charm_data=None, limit=None, import_filter=None):
    yield {'branch_dir': 'foo', 'branch_spec': 'bar'}


class TestLPJob(JobTestBase):

    def test_request(self):
        charm_data = [[u'~charmers/charms/precise/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]

        def available_mock(url, request):
            return json.dumps(charm_data)

        with HTTMock(available_mock):
            charms = list(available_charms())

        self.assertEqual([{
            'branch_spec': u'~charmers/charms/precise/someproject/trunk',
            'branch_deleted': False,
            'owner': u'charmers',
            'series': u'precise',
            'name': u'someproject',
            'bname': u'trunk',
            'commit': u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
            'distro_series': [],
            'promulgated': False}], charms)

    def test_request_promulgated_charm(self):
        charm_data = [[u'~charmers/charms/precise/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       ['precise']]]

        def available_mock(url, request):
            return json.dumps(charm_data)

        with HTTMock(available_mock):
            charms = list(available_charms())

        self.assertEqual([{
            'branch_spec': u'~charmers/charms/precise/someproject/trunk',
            'branch_deleted': False,
            'owner': u'charmers',
            'series': u'precise',
            'name': u'someproject',
            'bname': u'trunk',
            'commit': u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
            'distro_series': ['precise'],
            'promulgated': True}], charms)

    def test_available_charms_parse_failure(self):
        # If the charm cannot be parsed, it is not returned.
        handler = self.get_handler("charm.launchpad")
        charm_data = [[u'~charmers/charms/precise/junk/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        charms = [charm for charm in available_charms(charm_data)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Unable to parse repo "
                   "~charmers/charms/precise/junk/someproject/trunk")
        self.assertIn(err_msg, log_messages)
        self.assertEqual([], charms)

    def test_all_charms_filter(self):
        # If the branch doesn't start with the defined filter (i.e.
        # ~charmers/charms/precise) it is not returned
        charm_data = [[u'~not-charmers/charms/precise/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        factory.makeCharm(self.db, owner='not-charmers')
        with patch('charmworld.jobs.lp.CHARM_IMPORT_FILTER',
                   '~charmers/charms/precise'):
            charms = [charm for charm in all_charms(self.db, charm_data)]
        self.assertEqual([], charms)

    def test_all_charms_explicit_filter(self):
        # If the branch doesn't start with the defined filter (i.e.
        # ~charmers/charms/precise) it is not returned
        lp_data = [
            [u'~not-charmers/charms/precise/someproject/trunk',
             u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
             []],
            [u'~charmers/charms/precise/someproject/trunk',
             u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
             []],
        ]
        factory.makeCharm(
            self.db, owner='charmers', name='only-in-db')
        factory.makeCharm(
            self.db, owner='not-charmers', name='only-in-db')
        charms = all_charms(
            self.db, lp_data, import_filter='~charmers/charms/precise')
        self.assertEqual(['~charmers/charms/precise/only-in-db/trunk',
                          '~charmers/charms/precise/someproject/trunk'],
                         [charm['branch_spec'] for charm in charms])

    def test_available_charms_not_trunk(self):
        # If the branch is not trunk or trunk-1, the charm is not returned by
        # available charms.
        handler = self.get_handler("charm.launchpad")
        charm_data = [[u'~charmers/charms/precise/someproject/foo',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        charms = [charm for charm in available_charms(charm_data)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Skipped branch ~charmers/charms/precise/someproject/foo")
        self.assertIn(err_msg, log_messages)
        self.assertEqual([], charms)

    def test_all_charms_limit(self):
        handler = self.get_handler("charm.launchpad")
        charm_data = [
            [u'~charmers/charms/precise/someproject-%s/trunk' % count,
             u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc', []]
            for count in xrange(100)]
        for count in range(20):
            factory.makeCharm(self.db)
        charms = [charm for charm in all_charms(
            self.db, charm_data, limit=110)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Import limit reached (110)... stopping")
        self.assertIn(err_msg, log_messages)
        self.assertEqual(110, len(charms))

    def test_db_charms(self):
        charm_1 = factory.makeCharm(
            self.db, owner='o2', series='s1', name='c1')[1]
        charm_2 = factory.makeCharm(
            self.db, owner='o1', series='s2', name='c1')[1]
        charm_3 = factory.makeCharm(
            self.db, owner='o1', series='s1', name='c2')[1]
        charm_4 = factory.makeCharm(
            self.db, owner='o1', series='s1', name='c1')[1]
        result = [c['branch_spec'] for c in db_charms(self.db)]
        expected = [
            c['branch_spec'] for c in (charm_4, charm_3, charm_2, charm_1)]
        self.assertEqual(expected, result)

    def makeLPData(self, owner, series, name, promulgated=False):
        if promulgated:
            linked_series = [series]
        else:
            linked_series = []
        return ['~%s/charms/%s/%s/trunk' % (owner, series, name),
                '%s@example.com-%s-%s' % (owner, series, name), linked_series]

    def makeCharmAndLPData(self, owner, series, name, promulgated=False):
        charm = factory.makeCharm(
            self.db, owner=owner, series=series, name=name,
            promulgated=promulgated)[1]
        return charm, self.makeLPData(owner, series, name, promulgated)

    def test_all_charms(self):
        lp_data = []
        charm_1, lp_1 = self.makeCharmAndLPData(
            'owner', 'series', 'c1', promulgated=True)
        lp_data.append(lp_1)
        charm_2, lp_2 = self.makeCharmAndLPData(
            'owner', 'series', 'c2', promulgated=True)
        lp_3 = self.makeLPData('owner', 'series', 'c3')
        lp_data.append(lp_3)
        result = list(all_charms(self.db, lp_data))

        # Note that data for the promulgated charm_2 does not appear
        # in lp_data; db_charms() changed its promulgated status to
        # false.
        expected = [
            {
                'branch_spec': '~owner/charms/series/c1/trunk',
                'branch_deleted': False,
                'bname': 'trunk',
                'commit': 'owner@example.com-series-c1',
                'distro_series': ['series'],
                'name': 'c1',
                'owner': 'owner',
                'promulgated': True,
                'series': 'series',
            },
            {
                'branch_spec': '~owner/charms/series/c2/trunk',
                'branch_deleted': True,
                'bname': 'trunk',
                'distro_series': [],
                'name': 'c2',
                'owner': 'owner',
                'series': 'series',
                'promulgated': False,
                '_id': '~owner/charms/series/c2/trunk',
            },
            {
                'branch_spec': '~owner/charms/series/c3/trunk',
                'branch_deleted': False,
                'bname': 'trunk',
                'commit': 'owner@example.com-series-c3',
                'distro_series': [],
                'name': 'c3',
                'owner': 'owner',
                'promulgated': False,
                'series': 'series',
            },
        ]
        self.assertEqual(expected, result)

    @patch('charmworld.jobs.lp.available_charms', available_charms_mock)
    def test_queue_charms(self):
        out_queue = get_queue('test_queue')
        self.addCleanup(out_queue.clear)
        queue_charms(self.db, out_queue=out_queue)
        item = out_queue.next()
        self.assertEqual(
            {'branch_dir': 'foo', 'branch_spec': 'bar'}, item.payload)
