# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.jobs.ingest import BzrIngestJob
from charmworld.jobs.ingest import IngestError
from charmworld.jobs.ingest import ScanIngestJob
from charmworld.testing import factory
from charmworld.testing import JobTestBase
from charmworld.testing import REAL_FILE_CHARM_DIR
from charmworld.utils import quote_key


class TestScanJob(JobTestBase):

    def setUp(self):
        super(TestScanJob, self).setUp()
        self.bzr_job = BzrIngestJob()
        self.bzr_job.setup(db=self.db, root_dir=REAL_FILE_CHARM_DIR)
        self.scan_job = ScanIngestJob()
        self.scan_job.setup(db=self.db)
        ignore, self.charm = factory.makeCharm(self.db, with_real_files=True)
        self.charm['files'] = dict([
            (quote_key(cfile.filename), dict(cfile)) for cfile in
            self.bzr_job.store_branch_files(self.charm)
        ])

    def test_config_metadata_with_dotted_keys(self):
        del self.charm['config']
        self.scan_job.run(self.charm)
        # The key 'dotted.name' is replaced with 'dotted%2Ename', and
        # 'foo.bar' is replaced with 'foo%2Ebar'.
        self.assertFalse('dotted.name' in self.charm)
        self.assertEqual('foo', self.charm['dotted%2Ename'])
        self.assertFalse('foo.bar' in self.charm['config']['options'])
        self.assertEqual('baz', self.charm['config']['options']['foo%2Ebar'])

    def test_short_interface_specification(self):
        # Charms can specify the interfaces they require or provide
        # in two forms: The long form requires for each relation
        # a dictionary that has a least the key 'interface'; the short
        # form specifies only the interface name for as a relation's
        # value. (See
        # https://juju.ubuntu.com/docs/charm.html#the-metadata-file
        # for more details.)
        #
        # ScanJob.normalize_interfaces() transforms the short form
        # into the long form for further processing and for storage.
        del self.charm['provides']
        del self.charm['requires']
        self.scan_job.run(self.charm)
        # The sample metadata file has entries for the short and long
        # form of the interface specification in the "provides" and
        # in the "requires" section. The short entries are converted
        # into the long form.
        expected = {
            'database': {'interface': 'mongodb'},
            'sql_db': {'interface': 'pgsql'},
        }
        self.assertEqual(expected, self.charm['requires'])
        expected = {
            'ssl_website': {'interface': 'https'},
            'website': {'interface': 'http'},
        }
        self.assertEqual(expected, self.charm['provides'])

    def test_invalid_metadata_raises_IngestError(self):
        metadata = 'invalid_syntax: ['
        metadata_file = factory.make_charm_file(
            self.db, self.charm, '/path/to/metadata.yaml', content=metadata)
        self.charm['files'][quote_key('metadata.yaml')] = {
            'fileid': metadata_file._id
        }
        with self.assertRaises(IngestError) as ctx:
            self.scan_job.run(self.charm)
        info = str(ctx.exception)
        self.assertTrue(info.startswith(
            'Invalid charm metadata ~charmers/charms/precise/sample'))

    def test_invalid_config_yaml_raises_IngestError(self):
        config = 'invalid_syntax: ['
        config_file = factory.make_charm_file(
            self.db, self.charm, '/path/to/config.yaml', content=config)
        self.charm['files'][quote_key('config.yaml')] = {
            'fileid': config_file._id
        }
        with self.assertRaises(IngestError) as ctx:
            self.scan_job.run(self.charm)
        info = str(ctx.exception)
        self.assertTrue(info.startswith(
            'Invalid charm config yaml. ~charmers/charms/precise/sample'))

    def test_short_url_label_generation(self):
        # charm['short_url'] starts with the owner name for
        # unpromulgated charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': False,
        }
        self.scan_job.process_charm(charm)
        self.assertEqual('/~foo/precise/bar', charm['short_url'])
        # The label of unpromulgated charms starts with the owner name.
        self.assertEqual('~foo:precise/bar', charm['label'])
        # Promulgated charms do not have the owner name in the short URL
        # and in the label.
        charm = {
            'owner': 'charmers',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        self.scan_job.process_charm(charm)
        self.assertEqual('/charms/precise/bar', charm['short_url'])
        self.assertEqual('precise/bar', charm['label'])
        # This is also true if the charm is not owned by charmes.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        self.scan_job.process_charm(charm)
        self.assertEqual('/charms/precise/bar', charm['short_url'])
        self.assertEqual('precise/bar', charm['label'])
