# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch
from requests import Response

from charmworld.jobs.ingest import StoreIngestJob
from charmworld.jobs.ingest import addresses
from charmworld.testing import factory
from charmworld.testing import JobTestBase


class TestStoreJob(JobTestBase):

    def _mock_data(self, warning=False, error=False):
        data = {'revision': 1}
        if warning:
            data['warnings'] = 'Warning forced by mock.'
        if error:
            data['errors'] = 'Error forced by mock.'
        return data

    def falsify_session(self, job, content='[[0]]'):

        class FakeSession:

            @staticmethod
            def send(request):
                response = Response()
                response._content = content
                return response

        job.store.session = FakeSession

    def test_error_handling_error(self):
        handler = self.get_handler('charm.store')
        job = StoreIngestJob()
        job.setup()
        self.falsify_session(job)
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        with patch.object(job, '_store_get',
                          lambda x: self._mock_data(error=True)):
            job.run(charm)

        first_address = "cs:%s/%s" % (charm["series"], charm["name"])
        second_address = "cs:~%s/%s/%s" % (
            charm["owner"], charm["series"], charm["name"])
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = "rechecking %s with %s" % (first_address, second_address)
        self.assertIn(err_msg, log_messages)
        err_msg = ("store error on %s {'errors': 'Error forced by mock.',"
                   " 'revision': 1}" % second_address)
        self.assertIn(err_msg, log_messages)

    def test_error_handling_warning(self):
        handler = self.get_handler('charm.store')
        job = StoreIngestJob()
        job.setup()
        self.falsify_session(job)
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        with patch.object(job, '_store_get',
                          lambda x: self._mock_data(warning=True)):
            job.run(charm)

        first_address = "cs:%s/%s" % (charm["series"], charm["name"])
        second_address = "cs:~%s/%s/%s" % (
            charm["owner"], charm["series"], charm["name"])
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = "rechecking %s with %s" % (first_address, second_address)
        self.assertIn(err_msg, log_messages)
        err_msg = ("store error on %s {'warnings': 'Warning forced by mock.',"
                   " 'revision': 1}" % second_address)
        self.assertIn(err_msg, log_messages)

    def test_run(self):
        job = StoreIngestJob()
        job.setup()
        self.falsify_session(job, '[["2010-12-24", 4], ["2010-12-25", 1]]')
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        del charm['store_url']
        with patch.object(job, '_store_get', lambda x: self._mock_data()):
            job.run(charm)
        data = charm['store_data']
        self.assertIn('store_checked', data.keys())
        self.assertEqual(1, data['revision'])
        self.assertEqual(
            "cs:%s/%s-%d" % (charm["series"], charm["name"], 1),
            charm['store_url'])
        self.assertEqual(5, charm['downloads_in_past_30_days'])

    def test_addresses(self):
        # The helper function addresses() returns the cs:~owner variant
        # of a store URL for unpromulgated charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': False,
        }
        self.assertEqual(['cs:~foo/precise/bar'], list(addresses(charm)))
        # addresses() returns the URL and the long URL for promulgated
        # charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        self.assertEqual(
            ['cs:precise/bar', 'cs:~foo/precise/bar'],
            list(addresses(charm)))
