# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import argparse
import logging
import sys
import time

from charmworld.utils import (
    configure_logging,
    get_ini,
)
from charmworld.jobs.config import CHARM_QUEUE
from charmworld.jobs.ingest import (
    run_job,
    UpdateCharmJob,
)
from charmworld.jobs.utils import get_queue
from charmworld.search import SearchServiceNotAvailable


DEFAULT_JOBS = (
    UpdateCharmJob,
)

settings = get_ini()
INTERVAL = int(settings.get('worker_interval'))


class CharmWorker(object):
    """Processes raw charm data"""

    def __init__(self, ingest_jobs, queue, interval=INTERVAL):
        self.ingest_jobs = ingest_jobs
        self.queue = queue
        self.log = logging.getLogger('charm.worker')
        self.interval = interval

    def run(self, quit_on_empty=False):
        self.log.info('Starting processing')
        return_code = 0
        for job in self.ingest_jobs:
            job.setup()
        while True:
            item = self.queue.next()
            if item is None:
                if quit_on_empty:
                    self.log.info(
                        'Empty queue with quit_on_empty set. Quitting.')
                    break
                else:
                    wait_time_in_minutes = self.interval / 60
                    self.log.info(
                        'Empty queue. Waiting %s minutes to recheck.' %
                        wait_time_in_minutes)
                    time.sleep(self.interval)
                    continue
            charm_data = item.payload
            try:
                for job in self.ingest_jobs:
                    try:
                        if not run_job(job, charm_data, needs_setup=False):
                            break
                    except SearchServiceNotAvailable as e:
                        self.log.exception(str(e))
                        self.queue.clear()
                        return_code = 1

            finally:
                item.complete()
        return return_code


# XXX j.c.sackett Feb 21 2013 Bug:1111708
# Functions not currently bound to jobs; probably need new jobs created for
# them outside of ingest worker.
#def reindex(db, indexer):
    #log = logging.getLogger("charm.index")
    #count = 0
    #for charm in db.find():
        #log.debug("Indexing %s", charm['branch_spec'])
        #try:
            #index_charm(indexer, charm)
        #except:
            #log.error("Indexing charm %s", charm['branch_spec'])
            #raise
        #count += 1
    #log.info("Indexed %d Charms" % count)


def main():
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--quit-on-empty",
        help="Shut down the worker if the queue is empty",
        action="store_true")
    # Explicitly pass sys.argv to make monkeypatching more reliable.
    args = parser.parse_args(sys.argv[1:])
    queue = get_queue(CHARM_QUEUE)
    jobs = [job() for job in DEFAULT_JOBS]
    worker = CharmWorker(jobs, queue)
    sys.exit(worker.run(quit_on_empty=args.quit_on_empty))


if __name__ == "__main__":
    main()
