# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import (
    date,
    datetime,
)
import urlparse

from requests import Response

from charmworld.charmstore import CharmStore
from charmworld.testing import (
    factory,
    TestCase,
)


class TestCharmStore(TestCase):

    def assertStoreURL(self, expected_path, url):
        self.assertEqual(CharmStore.STAT_URL + expected_path, url)

    def test_get_stats_range_request(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        start = date(2011, 01, 02)
        end = date(2012, 12, 31)
        cs = CharmStore()
        req = cs.get_stats_range_request(charm, start, end)
        self.assertEqual('GET', req.method)
        self.assertStoreURL('/charm-bundle:foo:bar:baz', req.url)
        self.assertEqual({
            'start': '2011-01-02',
            'end': '2012-12-31',
            'format': 'json',
            'by': 'day',
        }, req.params)
        startdatetime = datetime(2011, 02, 03)
        req = cs.get_stats_range_request(charm, startdatetime, startdatetime)
        self.assertEqual('2011-02-03', req.params['start'])
        self.assertEqual('2011-02-03', req.params['end'])
        charm['store_url'] = 'cs:foo/bar'
        url = cs.get_stats_range_request(charm, start, end).url
        self.assertEqual(
            'https://store.juju.ubuntu.com/stats/counter/charm-bundle:foo:bar',
            url)

    def test_get_download_counts(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        start = date(2011, 01, 02)
        end = date(2012, 12, 31)
        cs = CharmStore()

        class FakeSession:

            @staticmethod
            def send(request):
                parts = urlparse.urlsplit(request.url)
                self.assertEqual('/stats/counter/charm-bundle:foo:bar:baz',
                                 parts.path)
                query = urlparse.parse_qs(parts.query)
                self.assertEqual({
                    'start': ['2011-01-02'],
                    'end': ['2012-12-31'],
                    'by': ['day'],
                    'format': ['json']}, query)
                response = Response()
                response._content = '[[0]]'
                return response

        cs.session = FakeSession()
        result = cs.get_download_counts(charm, start, end)
        self.assertEqual([], result)

        class FakeSession2:
            @staticmethod
            def send(request):
                response = Response()
                response._content = '[["2013-15-34", 10]]'
                return response

        cs.session = FakeSession2()
        result = cs.get_download_counts(charm, start, end)
        self.assertEqual([["2013-15-34", 10]], result)

    def test_count_downloads_in_days(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        today = date(2011, 01, 02)
        cs = CharmStore()

        class FakeSession:

            @staticmethod
            def send(request):
                parts = urlparse.urlsplit(request.url)
                self.assertEqual('/stats/counter/charm-bundle:foo:bar:baz',
                                 parts.path)
                query = urlparse.parse_qs(parts.query)
                self.assertEqual(['2011-01-02'], query['end'])
                self.assertEqual(['2010-12-03'], query['start'])
                response = Response()
                response._content = '[["2010-12-24", 4], ["2010-12-25", 1]]'
                return response

        cs.session = FakeSession()

        self.assertEqual(5, cs.count_downloads_in_days(charm, 30, today))
