# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.jobs.ingest import BzrIngestJob
from charmworld.models import Charm
from charmworld.models import CharmSource
from charmworld.testing import factory
from charmworld.testing import login_request
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.utils import quote_key
from charmworld.views import SANITIZE
from charmworld.views.api import API2
from charmworld.views.charms import (
    _charm_view,
    charm_collection,
    config,
    distro_charm,
    distro_charm_json,
    hook,
    interface,
    featured_edit,
    personal_collection,
    personal_config,
    personal_hook,
    personal_series_collection,
    quality_edit,
    series_collection,
)
import json


class TestCharms(ViewTestBase):

    def test_charm_collection_basic(self):
        """Simple sanity check."""
        one_id, one = factory.makeCharm(self.db, promulgated=True)
        two_id, two = factory.makeCharm(self.db, promulgated=True, owner='foo')
        three_id, three = factory.makeCharm(self.db, promulgated=False)

        response = charm_collection(self.getRequest())
        charm_names = set(charm['name'] for charm in response['charms'])
        expected = set((one['name'], two['name']))
        self.assertEqual(expected, charm_names)

    def test_charm_collection_sub_filter(self):
        """With the subordinate filter we limit out the non-sub charms."""
        one_id, one = factory.makeCharm(
            self.db, subordinate=True, promulgated=True)
        two_id, two = factory.makeCharm(self.db, promulgated=True)

        request = self.getRequest()
        request.GET['sub'] = 'true'
        response = charm_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_charm_collection_charms_with_errors(self):
        """Charms with errors are not returned by charm_collection()."""
        one_id, one = factory.makeCharm(self.db, promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, charm_error=True, promulgated=True)
        response = charm_collection(self.getRequest())
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_personal_collection(self):
        # personal_collection() returns only charms owned by the given
        # person.
        one_id, one = factory.makeCharm(self.db, owner='foo')
        two_id, two = factory.makeCharm(self.db, owner='bar')
        request = self.getRequest()
        request.matchdict = {'owner': 'foo'}
        response = personal_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual('foo', charms[0]['owner'])

    def test_personal_collection_sub_filter(self):
        # If the request parameter 'sub' is set, only subordinate charms
        # are returned.
        one_id, one = factory.makeCharm(self.db, owner='foo')
        two_id, two = factory.makeCharm(self.db, owner='bar')
        three_id, three = factory.makeCharm(
            self.db, owner='foo', subordinate=True)
        request = self.getRequest()
        request.GET['sub'] = 'true'
        request.matchdict = {'owner': 'foo'}
        response = personal_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(three['name'], charms[0]['name'])
        # If the filter is no given in the request, all charms are returned.
        request = self.getRequest()
        request.matchdict = {'owner': 'foo'}
        response = personal_collection(request)
        charms = response['charms']
        self.assertEqual(2, len(charms))
        self.assertEqual('foo', charms[0]['owner'])
        self.assertEqual('foo', charms[1]['owner'])

    def test_personal_collection_charms_with_errors(self):
        # Charms with errors are not returned by personal_collection().
        one_id, one = factory.makeCharm(self.db, owner='foo')
        two_id, two = factory.makeCharm(self.db, owner='bar')
        three_id, three = factory.makeCharm(
            self.db, owner='foo', charm_error=True)
        request = self.getRequest()
        request.matchdict = {'owner': 'foo'}
        response = personal_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_series_collection(self):
        # series_collection() returns only promulgated charms for a
        # given series.
        one_id, one = factory.makeCharm(
            self.db, series='one', promulgated=True, name='charm-a')
        two_id, two = factory.makeCharm(
            self.db, series='two', promulgated=True)
        three_id, three = factory.makeCharm(
            self.db, series='one')
        four_id, four = factory.makeCharm(
            self.db, series='one', promulgated=True, owner='foo',
            name='charm-b')
        request = self.getRequest()
        request.matchdict = {'series': 'one'}
        response = series_collection(request)
        charms = response['charms']
        self.assertEqual(
            ['charm-a', 'charm-b'], [charm['name'] for charm in charms])
        self.assertEqual('one', charms[0]['series'])

    def test_series_collection_sub_filter(self):
        # If the request parameter 'sub' is set, only subordinate charms
        # are returned.
        one_id, one = factory.makeCharm(
            self.db, series='one', promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, series='two', promulgated=True)
        three_is, three = factory.makeCharm(
            self.db, series='one', subordinate=True, promulgated=True)
        request = self.getRequest()
        request.matchdict = {'series': 'one'}
        request.GET['sub'] = 'true'
        response = series_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(three['name'], charms[0]['name'])

    def test_series_collection_charms_with_errors(self):
        # Charms with errors are not returned by series_collection().
        one_id, one = factory.makeCharm(
            self.db, series='one', promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, series='two', promulgated=True)
        three_is, three = factory.makeCharm(
            self.db, series='one', charm_error=True, promulgated=True)
        request = self.getRequest()
        request.matchdict = {'series': 'one'}
        response = series_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_personal_series_collection(self):
        # personal_series_collection() returns only charms for a given
        # series and person.
        one_id, one = factory.makeCharm(self.db, series='one', owner='foo')
        two_id, two = factory.makeCharm(self.db, series='two', owner='foo')
        three_id, three = factory.makeCharm(self.db, series='one', owner='bar')
        request = self.getRequest()
        request.matchdict = {
            'owner': 'foo',
            'series': 'one',
        }
        response = personal_series_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_personal_series_collection_sub_filter(self):
        # If the parameter 'sub' is set, personal_series_collection()
        # returns only subordinate charms for the given person ans series.
        one_id, one = factory.makeCharm(self.db, series='one', owner='foo')
        two_id, two = factory.makeCharm(self.db, series='two', owner='foo')
        three_id, three = factory.makeCharm(self.db, series='one', owner='bar')
        four_id, four = factory.makeCharm(
            self.db, series='one', owner='foo', subordinate=True)
        request = self.getRequest()
        request.matchdict = {
            'owner': 'foo',
            'series': 'one',
        }
        request.GET['sub'] = 'true'
        response = personal_series_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(four['name'], charms[0]['name'])

    def test_personal_series_collection_charms_with_errors(self):
        # Charms with errors are not returned by series_collection().
        one_id, one = factory.makeCharm(self.db, series='one', owner='foo')
        two_id, two = factory.makeCharm(self.db, series='two', owner='foo')
        three_id, three = factory.makeCharm(self.db, series='one', owner='bar')
        four_id, four = factory.makeCharm(
            self.db, series='one', owner='foo', charm_error=True)
        request = self.getRequest()
        request.matchdict = {
            'owner': 'foo',
            'series': 'one',
        }
        response = personal_series_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])

    def test_interface(self):
        # interface() returns all charms reqiring or providing a given
        # interface.
        ignore, providing = factory.makeCharm(
            self.db, provides={'foo': {'interface': 'httpX'}})
        ignore, requiring = factory.makeCharm(
            self.db, requires={'bar': {'interface': 'httpX'}})
        ignore, other = factory.makeCharm(self.db)
        request = self.getRequest()
        request.matchdict = {'interface': 'httpX'}
        response = interface(request)
        self.assertEqual('httpX', response['interface'])
        found_providing = response['providers']
        self.assertEqual(1, len(found_providing))
        self.assertEqual(providing, found_providing[0])
        found_requiring = response['requirers']
        self.assertEqual(1, len(found_requiring))
        self.assertEqual(requiring, found_requiring[0])

    def test_interface_charm_with_errors(self):
        # Charms with error are not returned by interface().
        ignore, providing = factory.makeCharm(
            self.db, provides={'foo': {'interface': 'httpX'}})
        ignore, requiring = factory.makeCharm(
            self.db, requires={'bar': {'interface': 'httpX'}},
            charm_error=True)
        request = self.getRequest()
        request.matchdict = {'interface': 'httpX'}
        response = interface(request)
        self.assertEqual('httpX', response['interface'])
        found_providing = response['providers']
        self.assertEqual(1, len(found_providing))
        self.assertEqual(providing, found_providing[0])
        found_requiring = response['requirers']
        self.assertEqual(0, len(found_requiring))

    def test_config(self):
        ignore, charm = factory.makeCharm(
            self.db, name='foo', owner='bar', series='precise',
            promulgated=True)
        request = self.getRequest()
        request.matchdict = {'charm': 'foo', 'series': 'precise'}
        response = config(request)
        expected = {"charm": charm, "onload": "prettyPrint()"}
        self.assertEqual(expected, response)

    def test_personal_config(self):
        ignore, charm = factory.makeCharm(
            self.db, name='foo', owner='bar', series='precise',
            promulgated=False)
        request = self.getRequest()
        request.matchdict = {
            'charm': 'foo',
            'series': 'precise',
            'owner': 'bar'
        }
        response = personal_config(request)
        expected = {"charm": charm, "onload": "prettyPrint()"}
        self.assertEqual(expected, response)

    def make_charm_with_files(self, promulgated):
        ignore, charm = factory.makeCharm(
            self.db, owner='bar', series='precise',
            promulgated=promulgated, with_real_files=True)
        bzr_job = BzrIngestJob()
        bzr_job.setup(root_dir=None, db=self.db)
        bzr_job.add_files(charm)
        self.db.charms.update({'_id': charm['_id']}, charm)
        return charm

    def test_hook(self):
        charm = self.make_charm_with_files(promulgated=True)
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
            'hook': 'install',
        }
        response = hook(request)
        self.assertEqual('Not Set.', response['project'])
        self.assertEqual('install', response['hook'])
        self.assertTrue(response['hook_raw'].startswith(
            '#!/bin/bash\nset -eux\n\n# This PPA is no longer required.'))
        charm_data = dict(Charm(charm)._representation)
        self.assertEqual(charm_data, response['charm'])
        self.assertEqual('prettyPrint()', response['onload'])

    def test_personal_hook(self):
        charm = self.make_charm_with_files(promulgated=False)
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
            'hook': 'install',
            'owner': 'bar',
        }
        response = personal_hook(request)
        self.assertEqual('Not Set.', response['project'])
        self.assertEqual('install', response['hook'])
        self.assertTrue(response['hook_raw'].startswith(
            '#!/bin/bash\nset -eux\n\n# This PPA is no longer required.'))
        charm_data = dict(Charm(charm)._representation)
        self.assertEqual(charm_data, response['charm'])
        self.assertEqual('prettyPrint()', response['onload'])

    def test_distro_charm(self):
        self.enable_routes()
        self.make_charm_with_files(promulgated=True)
        request = self.getRequest()
        request.matchdict = {
            'charm': 'sample_charm',
            'series': 'precise',
        }
        response = distro_charm(request)
        expected_keys = set((
            'charm', 'format_change', 'format_proof', 'name', 'others',
            'project', 'qadata', 'readme', 'readme_format', 'icon_path'))
        self.assertEqual(expected_keys, set(response))

    def test_distro_charm_json(self):
        self.enable_routes()
        ignore, charm = factory.makeCharm(
            self.db, name='foo', owner='bar', series='precise',
            promulgated=True)
        request = self.getRequest()
        request.matchdict = {
            'charm': 'foo',
            'series': 'precise',
        }
        charm_data = dict(Charm(charm)._representation)
        for key in SANITIZE.intersection(charm_data.keys()):
            del charm_data[key]
        response = distro_charm_json(request)
        self.assertEqual(charm_data, json.loads(response.body))

    def test_valid_others(self):
        charm = factory.makeCharm(self.db, files={})[1]
        other_charm = factory.makeCharm(
            self.db, series=charm['series'], name=charm['name'],
            owner='jrandom')[1]
        response = _charm_view(self.getRequest(), charm)
        self.assertItemsEqual([other_charm], response['others'])
        other_charm['store_data']['errors'] = ['error']
        self.db.charms.save(other_charm)
        response = _charm_view(self.getRequest(), charm)
        self.assertItemsEqual([], response['others'])

    def test_charm_icons(self):
        self.enable_routes()
        no_icon_charm = factory.makeCharm(self.db, files={})[1]
        response = _charm_view(self.getRequest(), no_icon_charm)
        self.assertEqual(None, response['icon_path'])

        files = {quote_key('icon.svg'): {'filename': 'icon.svg'}}
        icon_charm = factory.makeCharm(self.db, files=files)[1]
        response = _charm_view(self.getRequest(), icon_charm)
        expect = '/api/2/charm/%s/file/icon.svg' % API2._get_api_id(
            Charm(icon_charm))
        self.assertEqual(expect, response['icon_path'])


class TestQAForm(ViewTestBase):

    def setUp(self):
        """Need to load the question data."""
        super(TestQAForm, self).setUp()
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        resp = quality_edit(request)

        self.assertIn('form', resp)
        self.assertIn('label', resp['form'])

    def test_charm_saves_data(self):
        """Verify good data causes the view to store qa data.

        This also verifies our schema a little bit as the fields are turned
        from strings in POST to int's per the schema to match the original
        factory generated data.

        """
        charm = factory.get_charm_json(promulgated=True)
        self.use_index_client()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        post_data = factory.makeQAAssessment()
        # Add the submit button to trigger the right path through the view.
        post_data['submit'] = 'submit'
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }

        quality_edit(request)

        # Remove the submit button we added since it's not stored into the
        # document.
        del post_data['submit']

        # Now fetching that charm should get us a charm with a qa attribute
        # matching the data submitted.
        for found in charm_source._get_all(charm['_id']):
            self.assertIn('qa', found)

            for key in post_data:
                # Ignore any form keys that start with __
                if key.startswith('__'):
                    continue

                # Ignore the deform fields that make up the nested structure
                # for the qa questions.
                if key.startswith('deform'):
                    continue

                self.assertEqual(
                    post_data[key],
                    found['qa'][key])

            # The mapped/nested data needs to exist in the updated charm data.
            self.assertIn('reliable', found['qa'])


class TestQAFormFunc(WebTestBase):
    """Functional tests for the QA Form Views."""

    def test_form_403s_for_non_charmers(self):
        """The QA form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db)
            # The status force the check against the status code.
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['series'], charm['name']),
                status=403)
            self.assertEqual(403, resp.status_code)

    def test_form_secured_from_users(self):
        """The qa form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['name'],
                charm['series']),
                status=200)
            self.assertEqual(resp.status_code, 200)


class TestFeaturedForm(ViewTestBase):

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        resp = featured_edit(request)

        self.assertIn('form', resp)
        self.assertIn('label', resp['form'])

    def test_form_saves_data(self):
        """Verify good data causes the view to store it."""
        charm = factory.get_charm_json(name='name1', promulgated=True)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'is_featured': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }

        self.assertFalse(charm['is_featured'])
        response = featured_edit(request)
        self.assertEqual(302, response.status_code)
        self.assertRegexpMatches(response.headers['Location'],
                                 'http://[^/]*/charms/precise/name1')
        for found in charm_source._get_all(charm['_id']):
            self.assertTrue(found['is_featured'])


class TestFeaturedFormFunc(WebTestBase):

    def test_form_403s_for_non_charmers(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db)
            # The status force the check against the status code.
            resp = resp.goto(
                'charms/%s/%s/featured/edit' %
                (charm['series'], charm['name']), status=403)
            self.assertEqual(403, resp.status_code)

    def test_form_secured_from_users(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            resp = resp.goto(
                'charms/%s/%s/featured/edit' %
                (charm['series'], charm['name']),
                status=200)
            self.assertEqual(resp.status_code, 200)
