# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.views.helpers import interfaces
from charmworld.testing import factory
from charmworld.testing import MongoTestBase


class TestHelpers(MongoTestBase):

    def test_interfaces(self):
        # view.helpers.interfaces() shows which interface are
        # available and which charms require and provide these
        # interfaces.
        one_id, one = factory.makeCharm(
            self.db, name='some_webapp',
            provides={'foo': {'interface': 'http'}},
            requires={'database': {'interface': 'pgsql'}},
            promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            promulgated=True)
        result = interfaces(self.db)
        expected = [
            {
                'name': 'http',
                'provided_by': [
                    {
                        '_id': '~charmers/charms/precise/some_webapp/trunk',
                        'i_provides': ['http'],
                        'i_requires': ['pgsql'],
                        'label': 'precise/some_webapp',
                        'name': 'some_webapp',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_webapp',
                        'store_url': 'cs:precise/some_webapp-1',
                    },
                ],
                'required_by': [],
            },
            {
                'name': 'pgsql',
                'provided_by': [
                    {
                        '_id': '~charmers/charms/precise/some_dbserver/trunk',
                        'i_provides': ['pgsql'],
                        'i_requires': [],
                        'label': 'precise/some_dbserver',
                        'name': 'some_dbserver',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_dbserver',
                        'store_url': 'cs:precise/some_dbserver-1',
                    },
                ],
                'required_by': [
                    {
                        '_id': '~charmers/charms/precise/some_webapp/trunk',
                        'i_provides': ['http'],
                        'i_requires': ['pgsql'],
                        'label': 'precise/some_webapp',
                        'name': 'some_webapp',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_webapp',
                        'store_url': 'cs:precise/some_webapp-1',
                    }
                ],
            }
        ]
        self.assertEqual(expected, result)

    def test_interfaces_does_not_list_invalid_charms(self):
        # Invalid charms are not returned by interfaces().
        one_id, one = factory.makeCharm(
            self.db, name='some_webapp',
            provides={'foo': {'interface': 'http'}},
            requires={'database': {'interface': 'pgsql'}},
            promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            charm_error=True, promulgated=True)
        result = interfaces(self.db)
        # Note that only the charm named "some_dbserver" is missing; the
        # entry for the interface "pgsql" is still listed because it is
        # required by the charm "webapp".
        expected = [
            {
                'name': 'http',
                'provided_by': [
                    {
                        '_id': '~charmers/charms/precise/some_webapp/trunk',
                        'i_provides': ['http'],
                        'i_requires': ['pgsql'],
                        'label': 'precise/some_webapp',
                        'name': 'some_webapp',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_webapp',
                        'store_url': 'cs:precise/some_webapp-1',
                    },
                ],
                'required_by': [],
            },
            {
                'name': 'pgsql',
                'provided_by': [],
                'required_by': [
                    {
                        '_id': '~charmers/charms/precise/some_webapp/trunk',
                        'i_provides': ['http'],
                        'i_requires': ['pgsql'],
                        'label': 'precise/some_webapp',
                        'name': 'some_webapp',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_webapp',
                        'store_url': 'cs:precise/some_webapp-1',
                    }
                ],
            }
        ]
        self.assertEqual(expected, result)

    def test_interfaces_unreviewed_charm(self):
        # Unreviewed charms are only returned if no reviewed charm
        # with the same name exists.
        one_id, one = factory.makeCharm(
            self.db, name='some_dbserver', owner='foo',
            provides={'database': {'interface': 'pgsql'}}, requires={})
        result = interfaces(self.db)
        expected = [
            {
                'name': 'pgsql',
                'provided_by': [
                    {
                        '_id': '~foo/charms/precise/some_dbserver/trunk',
                        'i_provides': ['pgsql'],
                        'i_requires': [],
                        'label': '~foo:precise/some_dbserver',
                        'name': 'some_dbserver',
                        'owner': 'foo',
                        'short_url': '/~foo/precise/some_dbserver',
                        'store_url': 'cs:~foo/precise/some_dbserver-1',
                    },
                ],
                'required_by': [],
            }
        ]
        self.assertEqual(expected, result)

        # If a reviewed charm with the same name exists, this charm is
        # listed instead.
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            promulgated=True)
        result = interfaces(self.db)
        expected = [
            {
                'name': 'pgsql',
                'provided_by': [
                    {
                        '_id': '~charmers/charms/precise/some_dbserver/trunk',
                        'i_provides': ['pgsql'],
                        'i_requires': [],
                        'label': 'precise/some_dbserver',
                        'name': 'some_dbserver',
                        'owner': 'charmers',
                        'short_url': '/charms/precise/some_dbserver',
                        'store_url': 'cs:precise/some_dbserver-1',
                    },
                ],
                'required_by': [],
            }
        ]
        self.assertEqual(expected, result)
